# -*- coding: utf-8 -*-
##############################################################################
#
#    wms module for OpenERP, WMS Extended for Open ERP
#    Copyright (C) 2010 SYLEAM Info Services (<http://www.syleam.fr/>)
#              Christophe CHAUVET <christophe.chauvet@syleam.fr>
#
#    This file is a part of wms
#
#    wms is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    wms is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv
from osv import fields
from tools.translate import _
import netsvc

_picking_state = [
    ('draft', 'Draft'),
    ('auto', 'Waiting'),
    ('confirmed', 'Confirmed'),
    ('assigned', 'Assigned'),
    ('done', 'Done'),
    ('cancel', 'Cancelled'),
]


class StockPicking(osv.osv):
    """
    Add crossdock process
    """
    _inherit = 'stock.picking'

    def _action_type(self, cr, uid, context=None):
        """
        Inherit properly the list of type
        """
        if context is None:
            context = {}
        res = super(StockPicking, self)._columns['type'].selection
        if 'transfer' not in [k for k, v in res]:
            res.append(('transfer', 'Transfer'))
        return res

    _columns = {
        'type': fields.selection(_action_type, 'Type', required=True),
        'warehouse_id': fields.many2one('stock.warehouse', 'Warehouse', help='Indicate the warehouse for this picking'),
        'picking_time': fields.float('Time', help='Time to made this picking'),
        'transporter_id': fields.many2one('res.partner', 'Transporter'),
        'transporter_number': fields.char('Tracking number', size=128),
        'product_id': fields.related('move_lines', 'product_id', type='many2one', relation='product.product', string='Product'),
        'round_id': fields.many2one('stock.round', 'Round', help='Use for manage deliveries'),
        'state': fields.selection(_picking_state, 'Status', readonly=True),
        'relay_address_id': fields.many2one('res.partner.address', 'Relay address', help='If the customer prefer to receive the delevery in a relay center, choose this address'),
        'delivery_amount': fields.float('Delivery amount', help='Fill the cost price for this delivery'),
        'delivery_amount_currency_id': fields.many2one('res.currency', 'Amount currency'),
    }

    _defaults = {
        'warehouse_id': lambda obj, cr, uid, context: obj.pool.get('res.users')._default_warehouse(cr, uid, context),
        'picking_time': lambda *a: 0.0,
        'transporter_id': lambda *a: False,
        'transporter_number': lambda *a: False,
        'round_id': lambda *a: False,
        'state': lambda *a: 'draft',
        'relay_address_id': lambda *a: False,
        'delivery_amount': lambda *a: 0.0,
        'delivery_amount_currency_id': lambda *a: False,
    }

    def default_get(self, cr, uid, fields_list, context=None):
        if context is None:
            context = {}
        values = super(StockPicking, self).default_get(cr, uid, fields_list, context=context)
        if 'type' in fields_list and values.get('type') == 'transfer' and values.get('name', '/') == '/':
            values['name'] = self.pool.get('ir.sequence').get(cr, uid, 'stock.picking.transfer')

        return values

    def onchange_wms(self, cr, uid, ids, address_id, context=None):
        """
        This on_change manage all values for WMS module
        """
        if context is None:
            context = {}

        vals = {}
        if address_id:
            addr = self.pool.get('res.partner.address').browse(cr, uid, address_id, context=context)
            if addr.round_id:
                vals['round_id'] = addr.round_id.id
            elif addr.partner_id and addr.partner_id.round_id:
                vals['round_id'] = addr.partner_id.round_id.id
        else:
            vals['round_id'] = False

        return {'value': vals}

    def create(self, cr, uid, values, context=None):
        """
        When picking is create, check if there is an address, and retrieve round value
        if round is on address, take it or on partner
        """
        if context is None:
            context = {}

        location_id = False
        if values.get('address_id') and not values.get('round_id', False) and not context.get('disable_round', False):
            addr = self.pool.get('res.partner.address').browse(cr, uid, values.get('address_id'), context=context)
            if addr.round_id:
                location_id = addr.round_id.default_location_id.id
                values['round_id'] = addr.round_id.id
            elif addr.partner_id and addr.partner_id.round_id:
                location_id = addr.partner_id.round_id.default_location_id.id
                values['round_id'] = addr.partner_id.round_id.id

        # Check if picking is create from picking out
        # and retrieve the transporter_id
        if values.get('type', '') == 'delivery' and values.get('name'):
            p_ids = self.search(cr, uid, [('type', '=', 'out'), ('name', '=', values['name'])], context=context)
            if p_ids and len(p_ids) == 1:
                transp = self.read(cr, uid, p_ids[0], ['transporter_id', 'transporter_weight', 'relay_address_id', 'delivery_amount', 'delivery_amount_currency_id'], context=context)
                if transp:
                    values['transporter_id'] = transp.get('transporter_id') and transp['transporter_id'][0] or False
                    values['transporter_weight'] = transp.get('transporter_weight', 0.0)
                    values['relay_address_id'] = transp.get('relay_address_id', False)
                    values['delivery_amount'] = transp.get('delivery_amount', 0.0)
                    values['delivery_amount_currency_id'] = transp.get('delivery_amount_currency_id', False)

        return super(StockPicking, self).create(cr, uid, values, context=context)

    def action_confirm(self, cr, uid, ids, context=None):
        """
        Change destination location on move, if there is a round on the picking
        """
        move_obj = self.pool.get('stock.move')
        pick = self.browse(cr, uid, ids, context=context)
        for p in pick:
            if p.type == 'out' and p.state == 'draft' and p.round_id and p.round_id.default_location_id:
                move_ids = [x.id for x in p.move_lines]
                move_obj.write(cr, uid, move_ids, {'location_dest_id': p.round_id.default_location_id.id}, context=context)

        return super(StockPicking, self).action_confirm(cr, uid, ids, context=context)

    def _ref_crossdock_id(self, cr, uid):
        """
        Return the reference for crossdock category
        """
        ir_mod_obj = self.pool.get('ir.model.data')
        ir_id = ir_mod_obj._get_id(cr, uid, 'wms_location', 'location_category_crossdock')
        return ir_mod_obj.read(cr, uid, ir_id, ['res_id'])['res_id']

    def product_crossdock_location(self, cr, uid, product_id, warehouse_id, context=None):
        """
        Search id
        """
        if context is None:
            context = {}

        cr_id = self._ref_crossdock_id(cr, uid)
        loc_ids = self.pool.get('stock.location').search(cr, uid, [('warehouse_id', '=', warehouse_id), ('usage', '=', 'internal'), ('categ_id', '=', cr_id)], context=context)
        move_ids = self.pool.get('wms.report.stock.available').search(cr, uid, [('product_id', '=', product_id), ('location_id', 'in', loc_ids)], context=context)
        return [x.location_id.id for x in self.pool.get('wms.report.stock.available').browse(cr, uid, move_ids, context=context)]

    def product_crossdock_quantity(self, cr, uid, product_id, warehouse_id, context=None):
        """
        Retrieve the quantity available for this product on his crossdock location
        """
        if context is None:
            context = {}

        res = 0.0
        cr_ids = self.product_crossdock_location(cr, uid, product_id, warehouse_id, context=context)
        if cr_ids:
            ctx = context.copy()
            ctx['states'] = ('assigned', 'done',)
            ctx['what'] = ('in', 'out')
            ctx['location'] = cr_ids[0]
            res = self.pool.get('product.product').get_product_available(cr, uid, [product_id], context=ctx)[product_id]

        return res

    def crossdock_location_available(self, cr, uid, warehouse_ids, context=None):
        """
        Return the list ids of empty crossdock location for these warehouse
        You can specify only_one in context to retrieve 1 location at the time

        :param warehouse_ids: list of the warehouse
        :type  warehouse_ids: list
        :return: dict with the list of empty crossdock location
        :rtype: dict
        """
        if context is None:
            context = {}

        st_av_obj = self.pool.get('wms.report.stock.available')
        res = {}
        for w in warehouse_ids:
            tmp_ids = []
            cr_id = self._ref_crossdock_id(cr, uid)
            l_ids = self.pool.get('stock.location').search(cr, uid,
                        [('warehouse_id', '=', w), ('usage', '=', 'internal'), ('categ_id', '=', cr_id)], context=context)
            for a in l_ids:
                if not st_av_obj.search(cr, uid, [('location_id', '=', a)]):
                    tmp_ids.append(a)
                if tmp_ids and context.get('only_one'):
                    break

            if not tmp_ids:
                default_crossdock = self.pool.get('stock.warehouse').read(cr, uid, w, ['crossdock_location_id'], context=context)['crossdock_location_id']
                if default_crossdock:
                    tmp_ids = [default_crossdock[0]]
            res[str(w)] = tmp_ids
        return res

    def create_crossdock_move(self, cr, uid, move, crossdock_id=None, context=None):
        """
        Check if the product on this stock.move have been reserved,
        and affecte the location on the picking_order

        :param move: movement to check
        :type  move: osv.osv.browse
        :return: True if the movement is correct
        :rtype: boolean
        """
        print '[I] create_crossdock_move'
        if context is None:
            context = {}

        move_obj = self.pool.get('stock.move')
        if crossdock_id is None:
            default_crossdock = self.pool.get('stock.warehouse').read(cr, uid, move.location_dest_id.warehouse_id.id,
                                        ['crossdock_location_id'], context=context)['crossdock_location_id']
            if not default_crossdock:
                raise osv.except_osv(_('Error'), _('No default crossdock location defined on the warehouse!'))
            crossdock_id = isinstance(default_crossdock, tuple) and default_crossdock[0] or default_crossdock

        ###
        ## If we have an outgoing quantity, and location type is crossdock, we split in two line
        ##
        new_id = False
        print 'produit: %s' % move.product_id.name
        print 'quantite manquante %.2f' % move.product_id.outgoing_qty
        #if abs(move.product_id.outgoing_qty) > 0:

        ###
        ## Search how many quantity have been reserved, and move all of them to the crossdock location
        ##
        picking_ids = self.search(cr, uid, [('type', '=', 'out'), ('state', 'in', ['confirmed', 'assigned'])],
                            context=context)
        args = [
            ('picking_id', 'in', picking_ids),
            ('product_id', '=', move.product_id.id),
            ('state', 'in', ['assigned', 'confirmed', 'waiting']),
            ('crossdock_ok', '=', False),
        ]
        reserve_ids = move_obj.search(cr, uid, args, order='date_planned', context=context)
        print 'reserve_ids ', reserve_ids

        if reserve_ids:
            reserves = move_obj.browse(cr, uid, reserve_ids, context=context)

            ###
            ## Check if this product have available quantity in te crossdock location
            ##
            cross_qty = self.product_crossdock_quantity(cr, uid, move.product_id.id, move.picking_id.warehouse_id.id)
            print 'cross_qty ', cross_qty

            total_ids = []
            total_crossdock = 0.0
            partial_ids = []
            partial_crossdock = 0.0
            for r in reserves:
                res_qty = r.product_qty
                if cross_qty >= res_qty:
                    print 'cross_qty >= res_qty (%.2f >= %.2f)' % (cross_qty, res_qty)
                    cross_qty = cross_qty - res_qty
                    res_qty = 0.0
                else:
                    print 'cross_qty < res_qty (%.2f < %.2f)' % (cross_qty, res_qty)
                    res_qty = res_qty - cross_qty
                    cross_qty = 0.0

                print '------'
                print 'total_crossdock ', total_crossdock
                print 'r.product_qty ', r.product_qty
                print 'move.product_qty ', move.product_qty
                print 'res_qty ', res_qty
                print '(total_crossdock + res_qty) > move.product_qty (%.2f + %.2f) > %.2f' % (total_crossdock, res_qty, move.product_qty)
                if (total_crossdock + res_qty) > move.product_qty:
                    print 'a pas tout, donc partiel'
                    partial_ids.append(r.id)
                    partial_crossdock += res_qty
                    break
                total_ids.append(r.id)
                total_crossdock += res_qty

            print '==================================='
            print 'move.product_qty ', move.product_qty
            print 'total_ids ', total_ids
            print 'total_crossdock ', total_crossdock
            print 'partial_ids ', partial_ids
            print 'partial_crossdock ', partial_crossdock
            print '==================================='

            if partial_ids and move.picking_id.warehouse_id.force_reserved_location:
                ###
                ## if partial line is force to reserved, we move all quantity to the crossdock location
                ##
                args = {
                    'location_dest_id': crossdock_id,
                    'state': 'done',
                    'crossdock_ok': True,
                }
                move_obj.write(cr, uid, [move.id], args, context=context)

            elif total_crossdock and move.product_qty > total_crossdock:
                ###
                ## If we have more quantity that reversed one, we split the move line in two
                ##
                print '[ENOUGTH] ************************'
                rest = (move.product_qty - total_crossdock)
                args = {
                    'product_qty': rest,
                    'state': 'done',
                    'crossdock_ok': True,
                    'move_dest_id': False,
                }
                new_id = move_obj.copy(cr, uid, move.id, args, context=context)
                args = {
                    'product_qty': total_crossdock,
                    'location_dest_id': crossdock_id,
                    'state': 'done',
                    'crossdock_ok': True,
                }
                move_obj.write(cr, uid, [move.id], args, context=context)

                # Check if move_dest_id is link to an internal picking
                if move.move_dest_id and move.move_dest_id.picking_id.type == 'internal':
                    print 'On rectifie les quantites du bon de rangement'
                    move_obj.write(cr, uid, [move.move_dest_id.id], {'product_qty': rest}, context=context)

            elif total_crossdock:
                # We don't have enougth quantity, how many moves can be assigned with quantity
                args = {
                    'location_dest_id': crossdock_id,
                    'state': 'done',
                    'crossdock_ok': True,
                }
                move_obj.write(cr, uid, [move.id], args, context=context)

            ###
            ## made all movement on out picking available
            ##
            if total_ids:
                args = {
                    'location_id': crossdock_id,
                    'state': 'assigned',
                    'crossdock_ok': True,
                }
                move_obj.write(cr, uid, total_ids, args, context=context)

            ###
            ## If checkbox "force_reserved_location" move product to the reserved location
            ##
            if partial_ids and move.picking_id.warehouse_id.force_reserved_location:
                print 'pas assez mais reserver'
                args = {
                    'location_id': crossdock_id,
                }
                move_obj.write(cr, uid, partial_ids, args, context=context)

        print '[O] create_crossdock_move'
        return new_id or move.id

    def affect_reserved_product(self, cr, uid, move_id, location_id=False, context=None):
        """
        When product is reserved, we must affect receipt quantity to this one

        :param move_id: Movement ID
        :type  move_id: integer
        :param location_id: Location id for the product of this movement
        :type  location_id: integer
        :return: The quantity reserved by the picking out
        :rtype: float
        """
        print '[I] affect_reserved_product'
        if context is None:
            context = {}

        move_obj = self.pool.get('stock.move')
        move = move_obj.browse(cr, uid, move_id, context=context)

        ###
        ## If movement is in waiting, another partial receipt order have been received
        ##
        count_move = 0.0
        if move.move_dest_id.state in ('draft', 'cancel'):
            return False

        elif move.move_dest_id.state == 'waiting':
            args = [
                ('move_dest_id', '=', move.move_dest_id.id),
                ('id', '!=', move_id),
                ('state', '=', 'done'),
            ]
            old_move_ids = move_obj.search(cr, uid, args, context=context)
            if not old_move_ids:
                return False

            for om in move_obj.browse(cr, uid, old_move_ids, context=context):
                count_move += om.product_qty

        # If not location for crossdock, use the default on the warehouse
        if not location_id:
            location_id = move.location_dest_id.warehouse_id.crossdock_location_id.id

        vals = {}
        if move.move_dest_id.product_qty < (move.product_qty + count_move):
            vals['state'] = 'waiting'
        else:
            vals['state'] = 'assigned'
            vals['location_id'] = location_id

        # Write on link, if product is available or not
        move_obj.write(cr, uid, [move.move_dest_id.id], vals, context=context)

        # On the receipt link, change the location_dest_id to replace by the available one
        move_obj.write(cr, uid, [move_id], {'location_dest_id': location_id}, context=context)

        print '[O] affect_reserved_product'
        return move.move_dest_id.product_qty

    def test_finnished(self, cr, uid, ids):
        """
        On receipt order, check if there is picking, and generate crossdock
        1) check if all move are finnished
        2) if not return False
        """
        print '[I] test_finished ************************************************'
        in_ids = []
        other_ids = []
        for p in self.browse(cr, uid, ids):
            if p.type == 'in':
                in_ids.append(p.id)
            else:
                other_ids.append(p.id)

        # check if there is a receipt order, if not call directly the super method
        if len(in_ids) == 0:
            return super(StockPicking, self).test_finnished(cr, uid, ids)

        # if we have a receipt we must check if product have been reserved
        elif len(in_ids) >= 1:
            move_obj = self.pool.get('stock.move')
            lines_ok = True
            move_ids = move_obj.search(cr, uid, [('picking_id', 'in', in_ids)])
            cds_ids = []
            new_move_ids = []
            moves = move_obj.browse(cr, uid, move_ids)
            for move in moves:
                if not cds_ids and move.location_dest_id.warehouse_id.id:
                    cd_ids = self.crossdock_location_available(cr, uid, [move.location_dest_id.warehouse_id.id])
                    if isinstance(cd_ids, dict):
                        cd_ids = cd_ids[str(move.location_dest_id.warehouse_id.id)]

                if move.state not in ('done', 'cancel'):
                    if move.product_qty != 0.0:
                        lines_ok = False

            if not lines_ok:
                return False

            for move in moves:
                # If product is reserved to one picking, check it first
                # if not, check if product is in crossdock mode and if out picking is available
                if move.move_dest_id and move.move_dest_id.picking_id and move.move_dest_id.picking_id.type == 'out':
                    print 'This product is reserved'
                    pro_crossdock = self.product_crossdock_location(cr, uid, move.product_id.id,
                                        move.location_dest_id.warehouse_id.id)
                    if pro_crossdock:
                        tmp_crossdock = pro_crossdock[0]
                    else:
                        tmp_crossdock = cd_ids and cd_ids.pop() or None

                    self.affect_reserved_product(cr, uid, move.id, tmp_crossdock)

                elif not move.crossdock_ok and move.product_id.location_type == 'crossdock' and move.product_qty > 0:
                    ###
                    ## Check if product is already assigned to a crossdock location
                    ## If not retrieve one crossodck location or use the default one on the warehouse
                    ##
                    pro_crossdock = self.product_crossdock_location(cr, uid, move.product_id.id,
                                        move.location_dest_id.warehouse_id.id)
                    if pro_crossdock:
                        tmp_crossdock = pro_crossdock[0]
                    else:
                        tmp_crossdock = cd_ids and cd_ids.pop() or None

                    m_id = self.create_crossdock_move(cr, uid, move, crossdock_id=tmp_crossdock)
                    new_move_ids.append(m_id)

            #raise osv.except_osv(_('Error'), _('Message'))
            #
            # For each picking out in confirm, check all lines and verify if it available
            # If yes, execute the workflow.
            args = [
                ('type', '=', 'out'),
                ('state', '=', 'confirmed'),
            ]
            p_ids = self.search(cr, uid, args)
            pick_list = []
            for p in self.browse(cr, uid, p_ids):
                l_ok = True
                for l in p.move_lines:
                    if l.state not in ('assigned', 'cancel'):
                        l_ok = False

                if l_ok:
                    pick_list.append(p.id)

            for i in pick_list:
                self.force_assign(cr, uid, [i])

        print '[O] test_finished'
        return super(StockPicking, self).test_finnished(cr, uid, ids)

    def move_merge(self, cr, uid, picking_id, context=None):
        """
        Merge move lines, keep lot, tracking
        Returns move lines in dico : {'product_key': (quantity, move_line_id)}
        used by burst mode
        """
        if context is None:
            context = {}

        move_obj = self.pool.get('stock.move')
        move_ids = move_obj.search(cr, uid, [('picking_id', '=', picking_id)], context=context)
        known_moves = {}
        moves_to_delete = []
        for move in move_obj.browse(cr, uid, move_ids, context=context):
            product_key = "%d_%d_%d" % (move.product_id.id, move.prodlot_id and move.prodlot_id.id or 0,
                                        move.tracking_id and move.tracking_id.id or 0)
            # Check if product_id is known
            if product_key in known_moves.keys():
                # Add quantities
                known_moves[product_key].product_qty += move.product_qty
                moves_to_delete.append(move.id)
            else:
                known_moves[product_key] = move

        # Saves modified moves
        for p, move in known_moves.items():
            move_obj.write(cr, uid, [move.id], {'product_qty': move.product_qty}, context=context)
            known_moves[p] = (move.product_qty, move.id, )

        # Delete moves to delete
        if moves_to_delete:
            ctx = context.copy()
            ctx['call_unlink'] = True
            move_obj.unlink(cr, uid, moves_to_delete, context=ctx)

        return known_moves

    def reinit_draft(self, cr, uid, ids, context=None):
        """
        When picking is confirmed, we can put in draft,
        and delete all moves before reimport it
        """
        if context is None:
            context = {}

        wf_service = netsvc.LocalService("workflow")
        move_obj = self.pool.get('stock.move')
        ctx = context.copy()
        ctx.update({'call_unlink': True})
        for p in self.browse(cr, uid, ids, context=context):
            if p.state != 'confirmed':
                raise osv.except_osv(_('Error'), _('Only picking in confirmed or assigned must be reset!'))

            picking_link = []
            move_link = []
            move_current = []
            ## Retrieve all picking link to the movement
            for m in p.move_lines:
                move_current.append(m.id)
                if m.move_dest_id and m.move_dest_id.picking_id:
                    move_link.append(m.move_dest_id.id)
                    if m.move_dest_id.picking_id.id not in picking_link:
                        picking_link.append(m.move_dest_id.picking_id.id)

            ## Delete movement for the relate picking
            if move_link:
                move_obj.unlink(cr, uid, move_link, context=ctx)

            ## Delete the related picking
            if picking_link:
                self.unlink(cr, uid, picking_link, context=ctx)

            if move_current:
                move_obj.write(cr, uid, move_current, {'state': 'draft'}, context=ctx)

            #self.write(cr, uid, p.id, {'state': 'draft'}, ctx)
            wf_service.trg_validate(uid, 'stock.picking', p.id, 'button_draft', cr)

            if move_current:
                move_obj.unlink(cr, uid, move_current, context=ctx)

        return True

    def action_draft(self, cr, uid, ids, context=None):
        """
        move from cancel state to draft
        """
        if context is None:
            context = {}

        print 'action_draft ', ids
        self.write(cr, uid, ids, {'state': 'draft'}, context=context)
        return True

    def draft_validate(self, cr, uid, ids, context=None):
        """Force assignation iff not internal and complete the moves.

        If the corresponding flag on warehouse is set, those pickings in the
        list that are internal go through assignment logic, in effect
        forbidding the source locations to go negative.
        The other are forced."""

        self.action_confirm(cr, uid, ids, context=context)
        # dispatch
        regular = []
        forced = []
        wh_obj = self.pool.get('stock.warehouse')
        flag = 'forbid_unsupplied_internal_move'
        for item in self.read(cr, uid, ids, ['type', 'warehouse_id'],
                              context=context):
            whid = item['warehouse_id'][0]
            if item['type'] == 'internal' and whid and wh_obj.read(
                cr, uid, whid, [flag]):
                regular.append(item['id'])
            else:
                forced.append(item['id'])

        self.action_assign(cr, uid, regular)
        self.draft_force_assign(cr, uid, forced)
        for pick in self.browse(cr, uid, forced):
            move_ids = [x.id for x in pick.move_lines]
            self.pool.get('stock.move').force_assign(cr, uid, move_ids)


        # finish assigned moves and apply wkf logic
        wf_service = netsvc.LocalService("workflow")
        for pid in ids:
            wf_service.trg_write(uid, 'stock.picking', pid, cr)

        self.action_move(cr, uid, ids, context=context)

        for pid in ids:
            wf_service.trg_validate(uid, 'stock.picking',
                                    pid, 'button_done', cr)
        return True

    #def test_cancel(self, cr, uid, ids, context={}):
    #    """
    #    Redefine test cancel, to protect
    #    """
    #    for pick in self.browse(cr, uid, ids, context=context):
    #        for move in pick.move_lines:
    #            if move.state not in ('cancel',):
    #                return False
    #    return True

StockPicking()

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
